from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Related(FieldSet):

    def __init__(self,
                 ip: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.ip = ip


class RelatedSchema(FieldSetSchema):
    ip = fields.String()




from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class OS(FieldSet):

    def __init__(self,
                 family: str = None,
                 full: str = None,
                 kernel: str = None,
                 name: str = None,
                 platform: str = None,
                 version: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.family = family
        self.full = full
        self.kernel = kernel
        self.name = name
        self.platform = platform
        self.version = version


class OSSchema(FieldSetSchema):
    family = fields.String()
    full = fields.String()
    kernel = fields.String()
    name = fields.String()
    platform = fields.String()
    version = fields.String()



from datetime import datetime

from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Event(FieldSet):

    def __init__(self,
                 action: str = None,
                 category: str = None,
                 created: datetime = None,
                 dataset: str = None,
                 duration: int = None,
                 end: datetime = None,
                 hash: str = None,
                 id: str = None,
                 kind: str = None,
                 module: str = None,
                 original: str = None,
                 outcome: str = None,
                 risk_score: float = None,
                 risk_score_norm: float = None,
                 severity: int = None,
                 start: datetime = None,
                 timezone: str = None,
                 type: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = action
        self.category = category
        self.created = created
        self.dataset = dataset
        self.duration = duration
        self.end = end
        self.hash = hash
        self.id = id
        self.kind = kind
        self.module = module
        self.original = original
        self.outcome = outcome
        self.risk_score = risk_score
        self.risk_score_norm = risk_score_norm
        self.severity = severity
        self.start = start
        self.timezone = timezone
        self.type = type


class EventSchema(FieldSetSchema):
    action = fields.String()
    category = fields.String()
    created = fields.DateTime(format="iso")
    dataset = fields.String()
    duration = fields.Integer()
    end = fields.DateTime(format="iso")
    hash = fields.String()
    id = fields.String()
    kind = fields.String()
    module = fields.String()
    original = fields.String()
    outcome = fields.String()
    risk_score = fields.Float()
    risk_score_norm = fields.Float()
    severity = fields.Integer()
    start = fields.DateTime(format="iso")
    timezone = fields.String()
    type = fields.String()



from marshmallow import fields

from .fields import *

INCLUDE_FIELDS = {
    'agent': fields.Nested(AgentSchema),
    'client': fields.Nested(ClientSchema),
    'cloud': fields.Nested(CloudSchema),
    'container': fields.Nested(ContainerSchema),
    'destination': fields.Nested(DestinationSchema),
    'ecs': fields.Nested(ECSSchema),
    'error': fields.Nested(ErrorSchema),
    'event': fields.Nested(EventSchema),
    'file': fields.Nested(FileSchema),
    'geo': fields.Nested(GeoSchema),
    'group': fields.Nested(GroupSchema),
    'host': fields.Nested(HostSchema),
    'httprequest': fields.Nested(HttpRequestSchema),
    'httpresponse': fields.Nested(HttpResponseSchema),
    'logline': fields.Nested(LogLineSchema),
    'network': fields.Nested(NetworkSchema),
    'observer': fields.Nested(ObserverSchema),
    'organization': fields.Nested(OrganizationSchema),
    'os': fields.Nested(OSSchema),
    'process': fields.Nested(ProcessSchema),
    'related': fields.Nested(RelatedSchema),
    'server': fields.Nested(ServerSchema),
    'service': fields.Nested(ServiceSchema),
    'source': fields.Nested(SourceSchema),
    'url': fields.Nested(UrlSchema),
    'user': fields.Nested(UserSchema),
    'useragent': fields.Nested(UserAgentSchema),
}



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Container(FieldSet):

    def __init__(self,
                 id: str = None,
                 image_name: str = None,
                 image_tag: str = None,
                 labels: dict = None,
                 name: str = None,
                 runtime: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = id
        self.image_name = image_name
        self.image_tag = image_tag
        self.labels = labels
        self.name = name
        self.runtime = runtime


class ContainerSchema(FieldSetSchema):
    id = fields.String()
    image_name = fields.String()
    image_tag = fields.String()
    labels = fields.Dict()
    name = fields.String()
    runtime = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Observer(FieldSet):

    def __init__(self,
                 hostname: str = None,
                 ip: str = None,
                 mac: str = None,
                 serial_number: str = None,
                 type: str = None,
                 vendor: str = None,
                 version: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hostname = hostname
        self.ip = ip
        self.mac = mac
        self.serial_number = serial_number
        self.type = type
        self.vendor = vendor
        self.version = version


class ObserverSchema(FieldSetSchema):
    hostname = fields.String()
    ip = fields.String()
    mac = fields.String()
    serial_number = fields.String()
    type = fields.String()
    vendor = fields.String()
    version = fields.String()



from enum import Enum


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class Severity(OrderedEnum):
    CRITICAL = 5
    ERROR = 4
    WARNING = 3
    INFO = 2
    DEBUG = 1

    @staticmethod
    def from_str(severity_str: str) -> 'Severity':
        severity_str = severity_str.lower()
        if severity_str == "debug":
            return Severity.DEBUG
        elif severity_str == "info":
            return Severity.INFO
        elif severity_str == "warning":
            return Severity.WARNING
        elif severity_str == "error":
            return Severity.ERROR
        elif severity_str == "critical":
            return Severity.CRITICAL
        else:
            raise AssertionError(f"Invalid severity - {severity_str}")



from datetime import datetime
from typing import List

from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Process(FieldSet):

    def __init__(self,
                 args: List[str] = None,
                 executable: str = None,
                 name: str = None,
                 pid: int = None,
                 ppid: int = None,
                 start: datetime = None,
                 thread_id: int = None,
                 title: str = None,
                 working_directory: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.args = args
        self.executable = executable
        self.name = name
        self.pid = pid
        self.ppid = ppid
        self.start = start
        self.thread_id = thread_id
        self.title = title
        self.working_directory = working_directory


class ProcessSchema(FieldSetSchema):
    args = fields.List(fields.String())
    executable = fields.String()
    name = fields.String()
    pid = fields.Integer()
    ppid = fields.Integer()
    start = fields.DateTime(format="iso")
    thread_id = fields.Integer()
    title = fields.String()
    working_directory = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Client(FieldSet):

    def __init__(self,
                 address: str = None,
                 bytes: int = None,
                 domain: str = None,
                 ip: str = None,
                 mac: str = None,
                 packets: int = None,
                 port: int = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.address = address
        self.bytes = bytes
        self.domain = domain
        self.ip = ip
        self.mac = mac
        self.packets = packets
        self.port = port


class ClientSchema(FieldSetSchema):
    address = fields.String()
    bytes = fields.Integer()
    domain = fields.String()
    ip = fields.String()
    mac = fields.String()
    packets = fields.Integer()
    port = fields.Integer()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Cloud(FieldSet):

    def __init__(self,
                 account_id: str = None,
                 availability_zone: str = None,
                 instance_id: str = None,
                 instance_name: str = None,
                 machine_type: str = None,
                 provider: str = None,
                 region: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.account_id = account_id
        self.availability_zone = availability_zone
        self.instance_id = instance_id
        self.instance_name = instance_name
        self.machine_type = machine_type
        self.provider = provider
        self.region = region


class CloudSchema(FieldSetSchema):
    account_id = fields.String()
    availability_zone = fields.String()
    instance_id = fields.String()
    instance_name = fields.String()
    machine_type = fields.String()
    provider = fields.String()
    region = fields.String()



import json
import sys

from kubi_ecs_logger.models import Severity


def pprint(data: dict, output_destination=sys.stdout):
    """
    Prints the data in a formatted way with colors
    """
    severity = Severity.from_str(data["logline"]["level"])
    raw_data = ConsoleColors.get_color(severity) + json.dumps(data, indent=2, sort_keys=True) + ConsoleColors.ENDC
    print(raw_data, file=output_destination)


class ConsoleColors:
    GREY = '\033[37m'
    YELLOW = '\033[93m'
    CYAN = '\033[96m'
    RED = '\033[91m'
    DARK_RED = '\033[31m'
    ENDC = '\033[0m'

    @staticmethod
    def get_color(severity: Severity):
        if severity == Severity.DEBUG:
            return ConsoleColors.GREY
        if severity == Severity.INFO:
            return ConsoleColors.CYAN
        if severity == Severity.WARNING:
            return ConsoleColors.YELLOW
        if severity == Severity.ERROR:
            return ConsoleColors.RED
        if severity == Severity.CRITICAL:
            return ConsoleColors.DARK_RED



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class HttpRequest(FieldSet):

    def __init__(self,
                 body_bytes: int = None,
                 body_content: str = None,
                 bytes: int = None,
                 method: str = None,
                 referrer: str = None,
                 version: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.body_bytes = body_bytes
        self.body_content = body_content
        self.bytes = bytes
        self.method = method
        self.referrer = referrer
        self.version = version


class HttpRequestSchema(FieldSetSchema):
    body_bytes = fields.Integer()
    body_content = fields.String()
    bytes = fields.Integer()
    method = fields.String()
    referrer = fields.String()
    version = fields.String()


class HttpResponse(FieldSet):

    def __init__(self,
                 body_bytes: int = None,
                 body_content: str = None,
                 bytes: int = None,
                 status_code: str = None,
                 version: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.body_bytes = body_bytes
        self.body_content = body_content
        self.bytes = bytes
        self.status_code = status_code
        self.version = version


class HttpResponseSchema(FieldSetSchema):
    body_bytes = fields.Integer()
    body_content = fields.String()
    bytes = fields.Integer()
    status_code = fields.String()
    version = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Group(FieldSet):

    def __init__(self,
                 id: str = None,
                 name: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = id
        self.name = name


class GroupSchema(FieldSetSchema):
    id = fields.String()
    name = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class User(FieldSet):

    def __init__(self,
                 email: str = None,
                 full_name: str = None,
                 hash: str = None,
                 id: str = None,
                 name: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.email = email
        self.full_name = full_name
        self.hash = hash
        self.id = id
        self.name = name


class UserSchema(FieldSetSchema):
    email = fields.String()
    full_name = fields.String()
    hash = fields.String()
    id = fields.String()
    name = fields.String()




from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Geo(FieldSet):

    def __init__(self,
                 city_name: str = None,
                 continent_name: str = None,
                 country_iso_code: str = None,
                 country_name: str = None,
                 location: dict = None,
                 name: str = None,
                 region_iso_code: str = None,
                 region_name: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.city_name = city_name
        self.continent_name = continent_name
        self.country_iso_code = country_iso_code
        self.country_name = country_name
        self.location = location
        self.name = name
        self.region_iso_code = region_iso_code
        self.region_name = region_name


class GeoSchema(FieldSetSchema):
    city_name = fields.String()
    continent_name = fields.String()
    country_iso_code = fields.String()
    country_name = fields.String()
    location = fields.Dict()
    name = fields.String()
    region_iso_code = fields.String()
    region_name = fields.String()




from marshmallow import post_load
from kubi_ecs_logger.models import RootSchema


class FieldSet:

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)


class FieldSetSchema(RootSchema):
    class Meta:
        model = FieldSet

    @post_load
    def load_data(self, data):
        return FieldSet(**data)





from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Error(FieldSet):

    def __init__(self,
                 code: str = None,
                 id: str = None,
                 message: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code = code
        self.id = id
        self.message = message


class ErrorSchema(FieldSetSchema):
    code = fields.String()
    id = fields.String()
    message = fields.String()



"""
Logger is a wrapper around ECS-based logging models.
This class is initiated only ones and keeps the defaults and state.
When the class is called the base is reinitialized.
Also sending out the log will reinitialize the base.

Creating a log line:
Logger(message="This is my log message.").event(action="logging").out(severity='debug')

The first field object added since initialization of a specific type will be inserted,
the following field object of the same time will be neglected.
"""
import sys
from typing import Optional, List, Union
from datetime import datetime

from kubi_ecs_logger.models.fields import *

from kubi_ecs_logger.utils import pprint
from kubi_ecs_logger.models import Base, BaseSchema, Severity


class Logger:
    __instance: 'Logger' = None
    _defaults: dict = {}
    _base: Base = None
    _dev: bool = False
    _severity_output_level: Severity = Severity.INFO

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(Logger, cls).__new__(cls)
        cls.__instance.base()
        return cls.__instance

    @property
    def dev(self):
        return self._dev

    @dev.setter
    def dev(self, value):
        assert isinstance(value, bool)  # Check
        self._dev = value

    @property
    def severity_output_level(self):
        return self._severity_output_level

    @severity_output_level.setter
    def severity_output_level(self, value):
        assert isinstance(value, Severity)  # Check
        self._severity_output_level = value

    @property
    def defaults(self):
        return self._defaults

    @defaults.setter
    def defaults(self, value):
        assert isinstance(value, dict)  # Check
        self._defaults = value

    def base(self, date: datetime = None, labels: dict = None, message: str = None,
             tags: List[str] = None, **kwargs) -> 'Logger':
        defaults = self.__get_defaults_for(Base)
        if defaults:
            kwargs.update(defaults)

        self._base = Base(date=date, labels=labels, message=message, tags=tags, **kwargs)
        return self

    def agent(self, ephemeral_id: str = None, id: str = None, name: str = None,
              type: str = None, version: str = None, **kwargs):
        defaults = self.__get_defaults_for(Agent)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Agent(ephemeral_id=ephemeral_id, id=id, name=name, type=type, version=version, **kwargs))
        return self

    def client(self, address: str = None, bytes: int = None, domain: str = None, ip: str = None,
               mac: str = None, packets: int = None, port: int = None, **kwargs):
        defaults = self.__get_defaults_for(Client)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Client(address=address, bytes=bytes, domain=domain, ip=ip, mac=mac, packets=packets,
                                     port=port, **kwargs))
        return self

    def cloud(self, account_id: str = None, availability_zone: str = None, instance_id: str = None,
              instance_name: str = None, machine_type: str = None, provider: str = None,
              region: str = None, **kwargs):
        defaults = self.__get_defaults_for(Cloud)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Cloud(account_id=account_id, availability_zone=availability_zone, instance_id=instance_id,
                                    instance_name=instance_name, machine_type=machine_type, provider=provider,
                                    region=region, **kwargs))
        return self

    def container(self, id: str = None, image_name: str = None, image_tag: str = None,
                  labels: dict = None, name: str = None, runtime: str = None, **kwargs):
        defaults = self.__get_defaults_for(Container)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Container(id=id, image_name=image_name, image_tag=image_tag, labels=labels,
                                        name=name, runtime=runtime, **kwargs))
        return self

    def destination(self, address: str = None, bytes: int = None, domain: str = None, ip: str = None,
                    mac: str = None, packets: int = None, port: int = None, **kwargs):
        defaults = self.__get_defaults_for(Destination)
        if defaults:
            kwargs.update(defaults)
        self._base.add_object(Destination(address=address, bytes=bytes, domain=domain, ip=ip, mac=mac, packets=packets,
                                          port=port, **kwargs))
        return self

    def ecs(self, version: str = None, **kwargs):
        defaults = self.__get_defaults_for(Client)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(ECS(version=version))
        return self

    def error(self, code: str = None, id: str = None, message: str = None, **kwargs) -> 'Logger':
        defaults = self.__get_defaults_for(Error)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Error(code=code, id=id, message=message, **kwargs))
        return self

    def event(self, action: str = None, category: str = None, created: datetime = None,
              dataset: str = None, risk_score: float = None, severity: int = None,
              **kwargs) -> 'Logger':
        defaults = self.__get_defaults_for(Event)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Event(action=action, category=category, created=created, dataset=dataset,
                                    risk_score=risk_score, severity=severity, **kwargs))
        return self

    def file(self, ctime: datetime = None, device: str = None, extension: str = None, gid: str = None,
             group: str = None, inode: str = None, mode: str = None, mtime: datetime = None, owner: str = None,
             path: str = None, size: int = None, target_path: str = None, type: str = None, uid: str = None,
             **kwargs):
        defaults = self.__get_defaults_for(File)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(File(ctime=ctime, device=device, extension=extension, gid=gid, group=group,
                                   inode=inode, mode=mode, mtime=mtime, owner=owner, path=path, size=size,
                                   target_path=target_path, type=type, uid=uid, **kwargs))
        return self

    def geo(self, city_name: str = None, continent_name: str = None, country_iso_code: str = None,
            country_name: str = None, location: dict = None, name: str = None, region_iso_code: str = None,
            region_name: str = None, **kwargs):
        defaults = self.__get_defaults_for(Geo)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Geo(city_name=city_name, continent_name=continent_name, country_iso_code=country_iso_code,
                                  country_name=country_name, location=location, name=name,
                                  region_iso_code=region_iso_code,
                                  region_name=region_name, **kwargs))
        return self

    def group(self, id: str = None, name: str = None, **kwargs):
        defaults = self.__get_defaults_for(Group)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Group(id=id, name=name, **kwargs))
        return self

    def host(self, architecture: str = None, hostname: str = None, id: str = None, ip: str = None,
             mac: str = None, name: str = None, type: str = None, **kwargs):
        defaults = self.__get_defaults_for(Host)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Host(architecture=architecture, hostname=hostname, id=id, ip=ip, mac=mac,
                                   name=name, type=type, **kwargs))
        return self

    def http_request(self, body_bytes: int = None, body_content: str = None, bytes: int = None, method: str = None,
                     referrer: str = None, version: str = None, **kwargs):
        defaults = self.__get_defaults_for(HttpRequest)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(HttpRequest(body_bytes=body_bytes, body_content=body_content, bytes=bytes, method=method,
                                          referrer=referrer, version=version, **kwargs))
        return self

    def http_response(self, body_bytes: int = None, body_content: str = None, bytes: int = None,
                      status_code: str = None, version: str = None, **kwargs):
        defaults = self.__get_defaults_for(HttpResponse)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(HttpResponse(body_bytes=body_bytes, body_content=body_content, bytes=bytes,
                                           status_code=status_code, version=version, **kwargs))
        return self

    def log(self, level: Union[str, Severity] = None, original: str = None, **kwargs) -> 'Logger':
        defaults = self.__get_defaults_for(LogLine)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(LogLine(level=level, original=original, **kwargs))
        return self

    def network(self, application: str = None, bytes: int = None, community_id: str = None, direction: str = None,
                forwarded_ip: str = None, iana_number: str = None, name: str = None, packets: int = None,
                protocol: str = None, transport: str = None, type: str = None, **kwargs):
        defaults = self.__get_defaults_for(Network)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Network(application=application, bytes=bytes, community_id=community_id,
                                      direction=direction, forwarded_ip=forwarded_ip, iana_number=iana_number,
                                      name=name, packets=packets, protocol=protocol, transport=transport, type=type,
                                      **kwargs))
        return self

    def observer(self, hostname: str = None, ip: str = None, mac: str = None, serial_number: str = None,
                 type: str = None, vendor: str = None, version: str = None, **kwargs):
        defaults = self.__get_defaults_for(Observer)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Observer(hostname=hostname, ip=ip, mac=mac, serial_number=serial_number, type=type,
                                       vendor=vendor, version=version, **kwargs))
        return self

    def organization(self, id: str = None, name: str = None, **kwargs):
        defaults = self.__get_defaults_for(Organization)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Organization(id=id, name=name, **kwargs))
        return self

    def os(self, family: str = None, full: str = None, kernel: str = None, name: str = None, platform: str = None,
           version: str = None, **kwargs):
        defaults = self.__get_defaults_for(OS)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(OS(family=family, full=full, kernel=kernel, name=name, platform=platform, version=version,
                                 **kwargs))
        return self

    def process(self, args: List[str] = None, executable: str = None, name: str = None, pid: int = None,
                ppid: int = None, start: datetime = None, thread_id: int = None, title: str = None,
                working_directory: str = None, **kwargs):
        defaults = self.__get_defaults_for(Process)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Process(args=args, executable=executable, name=name, pid=pid, ppid=ppid, start=start,
                                      thread_id=thread_id, title=title, working_directory=working_directory, **kwargs))
        return self

    def related(self, ip: str = None, **kwargs):
        defaults = self.__get_defaults_for(Related)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Related(ip=ip, **kwargs))
        return self

    def server(self, address: str = None, bytes: int = None, domain: str = None, ip: str = None, mac: str = None,
               packets: int = None, port: int = None, **kwargs):
        defaults = self.__get_defaults_for(Server)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Server(address=address, bytes=bytes, domain=domain, ip=ip, mac=mac, packets=packets,
                                     port=port, **kwargs))
        return self

    def service(self, ephemeral_id: str = None, id: str = None, name: str = None, state: str = None, type: str = None,
                version: str = None, **kwargs):
        defaults = self.__get_defaults_for(Service)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Service(ephemeral_id=ephemeral_id, id=id, name=name, state=state, type=type,
                                      version=version, **kwargs))
        return self

    def source(self, address: str = None, bytes: int = None, domain: str = None, ip: str = None, mac: str = None,
               packets: int = None, port: int = None, **kwargs):
        defaults = self.__get_defaults_for(Source)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Source(address=address, bytes=bytes, domain=domain, ip=ip, mac=mac, packets=packets,
                                     port=port, **kwargs))
        return self

    def url(self, domain: str = None, fragment: str = None, full: str = None, original: str = None,
            password: str = None, path: str = None, port: int = None, query: str = None, scheme: str = None,
            username: str = None, **kwargs):
        defaults = self.__get_defaults_for(Url)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(Url(domain=domain, fragment=fragment, full=full, original=original, password=password,
                                  path=path, port=port, query=query, scheme=scheme, username=username, **kwargs))
        return self

    def user(self, email: str = None, full_name: str = None, hash: str = None, id: str = None, name: str = None,
             **kwargs):
        defaults = self.__get_defaults_for(User)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(User(email=email, full_name=full_name, hash=hash, id=id, name=name, **kwargs))
        return self

    def user_agent(self, device_name: str = None, name: str = None, original: str = None, version: str = None,
                   **kwargs):
        defaults = self.__get_defaults_for(UserAgent)
        if defaults:
            kwargs.update(defaults)

        self._base.add_object(UserAgent(device_name=device_name, name=name, original=original, version=version,
                                        **kwargs))
        return self

    def out(self, severity: Severity = Severity.DEBUG):
        assert isinstance(severity, Severity)

        self.__append_log_level(severity)

        if severity >= self._severity_output_level:
            self.__output()

        # Reset
        self.base()

    def __get_defaults_for(self, obj: type) -> Optional[dict]:
        name = str(obj.__name__).lower()
        if name in self._defaults:
            return self._defaults[name]
        return None

    def __output(self):
        """
        The internal function that handles/passes on the printing
        """
        if self._dev:
            pprint(BaseSchema().dump(self._base).data, output_destination=sys.stdout)
        else:
            print(BaseSchema().dumps(self._base).data, file=sys.stdout)

    def __append_log_level(self, severity_level: Severity):
        # Append log level if doesn't exist
        if hasattr(self._base, "logline"):
            if self._base.logline.level is None:
                self._base.logline.level = severity_level
            else:
                return

        self._base.add_object(LogLine(level=severity_level))



from datetime import datetime
from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class File(FieldSet):

    def __init__(self,
                 ctime: datetime = None,
                 device: str = None,
                 extension: str = None,
                 gid: str = None,
                 group: str = None,
                 inode: str = None,
                 mode: str = None,
                 mtime: datetime = None,
                 owner: str = None,
                 path: str = None,
                 size: int = None,
                 target_path: str = None,
                 type: str = None,
                 uid: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ctime = ctime
        self.device = device
        self.extension = extension
        self.gid = gid
        self.group = group
        self.inode = inode
        self.mode = mode
        self.mtime = mtime
        self.owner = owner
        self.path = path
        self.size = size
        self.target_path = target_path
        self.type = type
        self.uid = uid


class FileSchema(FieldSetSchema):
    ctime = fields.DateTime(format="iso")
    device = fields.String()
    extension = fields.String()
    gid = fields.String()
    group = fields.String()
    inode = fields.String()
    mode = fields.String()
    mtime = fields.DateTime(format="iso")
    owner = fields.String()
    path = fields.String()
    size = fields.Integer()
    target_path = fields.String()
    type = fields.String()
    uid = fields.String()




"""
TODO: Add doc what this file is doing
"""
from marshmallow import Schema, post_dump


class RootSchema(Schema):
    SKIP_VALUES = [None]

    @post_dump
    def remove_skip_values(self, data):
        return {
            key: value for key, value in data.items()
            if value not in self.SKIP_VALUES
        }

    @post_dump(pass_original=True)
    def add_extra(self, serialized, original):
        from kubi_ecs_logger.models.include import INCLUDE_FIELDS

        for k, v in original.__dict__.items():
            if k not in serialized and v is not None:
                type_name = str(type(v).__name__).lower()
                if type_name in INCLUDE_FIELDS:
                    schema = INCLUDE_FIELDS[type_name].schema
                    data = schema.dump(v).data
                    if "kind" not in data:
                        data["kind"] = type_name
                    serialized[k] = data
                elif isinstance(v, (int, float, str, bool, dict)):
                    if not str(k).startswith('_'):
                        serialized[k] = v

        return serialized



from typing import Union
from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema
from ..severity import Severity


class LogLine(FieldSet):

    def __init__(self,
                 level: Union[str, Severity] = None,
                 original: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        if isinstance(level, Severity):
            level = level.name

        self.__level = level
        self.original = original

    @property
    def level(self):
        return self.__level

    @level.setter
    def level(self, value):
        assert isinstance(value, Severity)
        self.__level = value.name


class LogLineSchema(FieldSetSchema):
    level = fields.String()
    original = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Host(FieldSet):

    def __init__(self,
                 architecture: str = None,
                 hostname: str = None,
                 id: str = None,
                 ip: str = None,
                 mac: str = None,
                 name: str = None,
                 type: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.architecture = architecture
        self.hostname = hostname
        self.id = id
        self.ip = ip
        self.mac = mac
        self.name = name
        self.type = type


class HostSchema(FieldSetSchema):
    architecture = fields.String()
    hostname = fields.String()
    id = fields.String()
    ip = fields.String()
    mac = fields.String()
    name = fields.String()
    type = fields.String()



from .field_set import FieldSet, FieldSetSchema
from .agent import Agent, AgentSchema
from .client import Client, ClientSchema
from .cloud import Cloud, CloudSchema
from .container import Container, ContainerSchema
from .destination import Destination, DestinationSchema
from .ecs import ECS, ECSSchema
from .error import Error, ErrorSchema
from .event import Event, EventSchema
from .file import File, FileSchema
from .geo import Geo, GeoSchema
from .group import Group, GroupSchema
from .host import Host, HostSchema
from .http import HttpRequest, HttpRequestSchema, HttpResponse, HttpResponseSchema
from .logline import LogLine, LogLineSchema
from .network import Network, NetworkSchema
from .observer import Observer, ObserverSchema
from .organization import Organization, OrganizationSchema
from .os import OS, OSSchema
from .process import Process, ProcessSchema
from .related import Related, RelatedSchema
from .server import Server, ServerSchema
from .service import Service, ServiceSchema
from .source import Source, SourceSchema
from .url import Url, UrlSchema
from .user import User, UserSchema
from .useragent import UserAgent, UserAgentSchema


from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class UserAgent(FieldSet):

    def __init__(self,
                 device_name: str = None,
                 name: str = None,
                 original: str = None,
                 version: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.device_name = device_name
        self.name = name
        self.original = original
        self.version = version

class UserAgentSchema(FieldSetSchema):
    device_name = fields.String()
    name = fields.String()
    original = fields.String()
    version = fields.String()





from typing import List
from datetime import datetime

from marshmallow import fields

from kubi_ecs_logger.models import RootSchema
from kubi_ecs_logger.models.fields import FieldSet


class Base:

    def __init__(self,
                 date: datetime = None,
                 labels: dict = None,
                 message: str = None,
                 tags: List[str] = None,
                 objects: List[FieldSet] = None,
                 **kwargs):
        self.timestamp = date or datetime.now()
        self.labels = labels
        self.message = message
        self.tags = tags

        # Add custom fields from **kwargs
        for k, v in kwargs.items():
            if not hasattr(self, k):
                setattr(self, k, v)

        # Add the field objects that are passed
        if objects is not None:
            for obj in objects:
                if not self.add_object(obj):
                    raise ValueError(f"You can have only one of: {type(obj).__name__}")

    # Add a field object after initialization
    def add_object(self, obj) -> bool:
        assert isinstance(obj, FieldSet)
        name = str(type(obj).__name__).lower()
        if not hasattr(self, name):
            setattr(self, name, obj)
            return True
        return False


class BaseSchema(RootSchema):
    class Meta:
        include = {
            "@timestamp": fields.DateTime(format="iso", attribute="timestamp", allow_none=False)
        }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from .include import INCLUDE_FIELDS
        self.declared_fields.update(INCLUDE_FIELDS)

    labels = fields.Dict(allow_none=False, skip_if=None)
    message = fields.String(allow_none=True)
    tags = fields.List(fields.String(), allow_none=True)



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Service(FieldSet):

    def __init__(self,
                 ephemeral_id: str = None,
                 id: str = None,
                 name: str = None,
                 state: str = None,
                 type: str = None,
                 version: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.ephemeral_id = ephemeral_id
        self.id = id
        self.name = name
        self.state = state
        self.type = type
        self.version = version


class ServiceSchema(FieldSetSchema):
    ephemeral_id = fields.String()
    id = fields.String()
    name = fields.String()
    state = fields.String()
    type = fields.String()
    version = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class ECS(FieldSet):

    def __init__(self,
                 version: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.version = version


class ECSSchema(FieldSetSchema):
    version = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Network(FieldSet):

    def __init__(self,
                 application: str = None,
                 bytes: int = None,
                 community_id: str = None,
                 direction: str = None,
                 forwarded_ip: str = None,
                 iana_number: str = None,
                 name: str = None,
                 packets: int = None,
                 protocol: str = None,
                 transport: str = None,
                 type: str = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.application = application
        self.bytes = bytes
        self.community_id = community_id
        self.direction = direction
        self.forwarded_ip = forwarded_ip
        self.iana_number = iana_number
        self.name = name
        self.packets = packets
        self.protocol = protocol
        self.transport = transport
        self.type = type


class NetworkSchema(FieldSetSchema):
    application = fields.String()
    bytes = fields.Integer()
    community_id = fields.String()
    direction = fields.String()
    forwarded_ip = fields.String()
    iana_number = fields.String()
    name = fields.String()
    packets = fields.Integer()
    protocol = fields.String()
    transport = fields.String()
    type = fields.String()



from marshmallow import fields

from .field_set import FieldSet, FieldSetSchema


class Url(FieldSet):

    def __init__(self,
                 domain: str = None,
                 fragment: str = None,
                 full: str = None,
                 original: str = None,
                 password: str = None,
                 path: str = None,
                 port: int = None,
                 query: str = None,
                 scheme: str = None,
                 username: str = None,
                 *aargs, **kwargs):
        super().__init__(*aargs, **kwargs)
        self.domain = domain
        self.fragment = fragment
        self.full = full
        self.original = original
        self.password = password
        self.path = path
        self.port = port
        self.query = query
        self.scheme = scheme
        self.username = username


class UrlSchema(FieldSetSchema):
    domain = fields.String()
    fragment = fields.String()
    full = fields.String()
    original = fields.String()
    password = fields.String()
    path = fields.String()
    port = fields.Integer()
    query = fields.String()
    scheme = fields.String()
    username = fields.String()
